from typing import Sequence, Dict, List
import numpy as np
from . import __version__
from .l1c_deconvolution_matrix_if import cimpl_create_deconvolution, cimpl_create_horizontal_stray_light_matrix, cimpl_set_numthreads, cimpl_set_diagnostic, cimpl_set_versionstr

WINDOW_UNDEFINED = 1
WINDOW_NOOF = 2
WINDOW_DETECTOR = 3
WINDOW_SCIENCE = 4
WINDOW_L1B = 5
WINDOW_L1C = 6


# ------------------------------------------------------------------------------
#           version_info
# ------------------------------------------------------------------------------
def deconvolution_matrix_version_info() -> str:
    return __version__


# ------------------------------------------------------------------------------
#           generate_convolution_matrix
# ------------------------------------------------------------------------------
def generate_convolution_matrix(array_window_specs: Sequence[int],
                                cell_measured_psf_arrays: Sequence[np.ndarray],
                                array_psf_center_locn: np.ndarray,
                                array_psf_interp_coeffs: np.ndarray,
                                array_psf_interp_indices: np.ndarray) -> np.ndarray:
    """
    Calculates the deconvolution matrix from an array of spatially varying point spread functions.
    measured at several locations across the surface.
    This code is numerically intensive and may take many hours to run for one wavelength/configuration.
    The function calls multi-threaded C++ code. The default operation will utilize all threads on
    the users system which can overwhelm the cooling system on some desktop machines.

    The generation of the L1C deconvolution matrix considers 5 areas on the detector surface.

    #. The Near-out-of-field window. Typically 6940 x 6940 pixels mapping to the 0.157 x 0.157 millirad near field of view
    #. The detector window. This is teh same size as the detector, 2000 x 1504
    #. The science winow. This is the useful scientific area of the detector window in detcetor pixels. Typically 1503 x 1500
    #. The L1B binned window. This window overlays the science window and bins the science pixels into L1B pixels
    #. The L1C binned window. This window overalsys the science and L1B windows and bins teh L1B pixels into L1C pixels.

    The following variables are used with the internal documentation,

    * Dx = Number of detector pixels in the X direction
    * Dy = Number of detector pixels in the Y direction
    * M  = Number of measured PSF arrays distributed across the detector surface
    * Ncoeff = Number of coefficients used to linearly interpolate the measured PSF's to any location on the detector surface
    * Kx = Number of L1C pixels in the X direction
    * Ky = Number of L1C pixels in the Y direction

    Parameters
    ----------
    array_window_specs : Array[18].
        Specify the placement and size of the five windows: NOOF, Detector, Science, L1B and L1C

        * [0] = **NOOFx**, X size of Near-Out-Of-Field window in detector pixels, e.g. 6940
        * [1] = **NOOFy**, Y size of Near-Out-Of-Field window in detector pixels, e.g. 6940
        * [2] = **DOFSx**, X offset of detector window in the NOOF window, e.g. 2718
        * [3] = **DOFSy**, Y offset of detector window in the NOOF window, e.g. 2718
        * [4] = **Dx**, X size of the detector window in detector pixels, e.g. 1504, Dx must equal psf_Nx
        * [5] = **Dy**, Y size of the detector window in detector pixels, e.g. 2000, Dy must equal psf_Ny
        * [6] = **SOFSx**, X offset of science window in the detector window in detector pixels, e.g. 0
        * [7] = **SOFSy**, Y offset of science window in the detector window in detector pixels, e.g. 248
        * [8] = **Sx**, X dimension of science window in detector pixels,  e.g. 1500, Sx must equal Lx*LBINx
        * [9] = **Sy**, Y dimension of science window in detector pixels,  e.g. 1503, sy must equal Ly*LBINy
        * [10] = **LBINx**, X L1B window pixel binning in detector pixels, e.g. 3
        * [11] = **LBINy**, Y L1B window pixel binning in detector pixels, e.g. 3
        * [12] = **Lx**, X number of L1B window pixels e.g. 500. Lx must equal Kx*KBINx
        * [13] = **Ly**, Y number of L1B window pixels e.g. 501. Ly must equal Ky*KBINy
        * [14] = **KBINx**, X L1C window pixel binning in L1B window pixels, e.g. 50
        * [15] = **KBINy**, Y L1C window pixel binning in L1B window pixels, e.g. 1
        * [16] = **Kx**, X number of L1C window pixels, e.g. 10
        * [17] = **Ky**, Y number of L1C window profile pixels, e.g. 501

    cell_measured_psf_arrays : List[psf arrays]
         A Sequence containing the M measured PSF arrays that will be interpolated
         to calculate the PSF at any other location in the detector useable area. Each element of the cell array
         is a measured PSF array expressed as a 2D array of size (Dy,Dx)

    array_psf_center_locn  : Array[M,3].
        A 2-D array containing two numbers for each of the M measured PSF arrays.
        The two numbers store the center location of the corresponding PSF array in detector coordinates as (Y,X, windowtype).

    array_psf_interp_coeffs  : Array(NCoeff, Dy, Dx).
        A 3-D array containing Ncoeffs coefficients used to interpolate the array of measured PSF arrays
        to any other location on the detector surface. Typically Ncoeffs will be ~3 to allow triangular interpolation
        (e.g. Delaunay Triangulation).

    array_psf_interp_indices : Array(NCoeff, Dy, Dx).
        A 3-D array containing NCoeff integers that index the M measured psf arrays
        in parameter cell_measured_psf_arrays. The indexed arrays are used for the linear interpolation.

    Returns
    -------
     np.ndarray
        The 2d convolution matrix of size Array( Ky*Kx, Ky*Kx). This array forward converts "true" data into the measured
        l1c blurred data. Deconvolution requires the inverse of this matrix.
    """

    window_specs = np.array(array_window_specs, dtype='int32', order='C')
    center_locn = np.array(array_psf_center_locn, dtype='float64', order='C')     # Make an array(M, 3)
    assert center_locn.shape[1] == 3
    assert window_specs.size == 22
    s = array_psf_interp_coeffs.shape
    NCoeffs = s[0]
    My = s[1]
    Mx = s[2]
    interp_indices = np.zeros([Mx, My, NCoeffs], dtype='int32', order='C')
    interp_coeffs = np.zeros([Mx, My, NCoeffs], dtype='float64', order='C')

    for i in range(NCoeffs):
        indices = array_psf_interp_indices[i, :, :].reshape([My, Mx])
        coeffs = array_psf_interp_coeffs[i, :, :].reshape([My, Mx])
        interp_indices[:, :, i] = indices.transpose()                                   # M arrays of column major arrays (Nx, Ny)
        interp_coeffs[:, :, i] = coeffs.transpose()                                     # M arrays of column major arrays (Nx, Ny)

    M = len(cell_measured_psf_arrays)                                                   # Get the number of PSF arrays in this list
    s = cell_measured_psf_arrays[0].shape                                               # Get the shape of each PSF array. They must be all the same
    Ny = s[0]                                                                           # Get the Number of rows (Y)
    Nx = s[1]                                                                           # Get the number of columns (X)
    psf_arrays = np.zeros([M, Nx, Ny], dtype='float64', order='C')                      # Create the array for C++ which uses the Column Major transpose
    for i in range(M):                                                                  # and copy
        psf_arrays[i, :, :] = cell_measured_psf_arrays[i].transpose()                   # the transpose for each array over to the system in column major format (Dx, Dy)

    Kx = window_specs[20]
    Ky = window_specs[21]
    A = np.zeros([Kx * Ky, Kx * Ky], dtype='float64', order='C')                        # Create the output array in the column major format
    cimpl_set_versionstr(__version__)
    cimpl_create_deconvolution(window_specs, psf_arrays, center_locn, interp_coeffs, interp_indices, A)
    A = np.array(A.transpose())                                                         # Make a copy back into the row major format (Ny,Nx)
    return A


# ------------------------------------------------------------------------------
#           generate_horizontal_stray_light_matrix
# ------------------------------------------------------------------------------
def generate_horizontal_stray_light_matrix(array_window_specs: Sequence[int],
                                           vignetting: np.ndarray,
                                           delta_steps: Dict[str, int],
                                           cell_measured_psf_arrays: Sequence[np.ndarray],
                                           array_psf_center_locn: np.ndarray,
                                           array_psf_interp_coeffs: np.ndarray,
                                           array_psf_interp_indices: np.ndarray) -> np.ndarray:
    """
    Calculates the deconvolution matrix for the horizontal stray light in the near out of field regions surrounding the
    left and right sides of the science window. This code is numerically intensive and may take many hours to run for
    one wavelength/configuration. The function calls multi-threaded C++ code. The default operation will utilize all threads on
    the users system which can overwhelm the cooling system on some desktop machines.

    Parameters
    ----------
    array_window_specs : Array[18].
        Specify the placement and size of the five windows: NOOF, Detector, Science, L1B and L1C

        * [0] = **NOOFx**, X size of Near-Out-Of-Field window in detector pixels, e.g. 6940
        * [1] = **NOOFy**, Y size of Near-Out-Of-Field window in detector pixels, e.g. 6940
        * [2] = **DOFSx**, X offset of detector window in the NOOF window, e.g. 2718
        * [3] = **DOFSy**, Y offset of detector window in the NOOF window, e.g. 2718
        * [4] = **Dx**, X size of the detector window in detector pixels, e.g. 1504, Dx must equal psf_Nx
        * [5] = **Dy**, Y size of the detector window in detector pixels, e.g. 2000, Dy must equal psf_Ny
        * [6] = **SOFSx**, X offset of science window in the detector window in detector pixels, e.g. 0
        * [7] = **SOFSy**, Y offset of science window in the detector window in detector pixels, e.g. 248
        * [8] = **Sx**, X dimension of science window in detector pixels,  e.g. 1500, Sx must equal Lx*LBINx
        * [9] = **Sy**, Y dimension of science window in detector pixels,  e.g. 1503, sy must equal Ly*LBINy
        * [10] = **LBINx**, X L1B window pixel binning in detector pixels, e.g. 3
        * [11] = **LBINy**, Y L1B window pixel binning in detector pixels, e.g. 3
        * [12] = **Lx**, X number of L1B window pixels e.g. 500. Lx must equal Kx*KBINx
        * [13] = **Ly**, Y number of L1B window pixels e.g. 501. Ly must equal Ky*KBINy
        * [14] = **KBINx**, X L1C window pixel binning in L1B window pixels, e.g. 50
        * [15] = **KBINy**, Y L1C window pixel binning in L1B window pixels, e.g. 1
        * [16] = **Kx**, X number of L1C window pixels, e.g. 10
        * [17] = **Ky**, Y number of L1C window profile pixels, e.g. 501

    vignetting: Array[NOOFy,NOOFx]
        The vignetting used for the near out of field regions. It can be set to a size of [0,0] in which case all
        vignetting is disabled and set to 1.0. Otherwise the vignetting array must be the size of the NOOF window.
        The vignetting value is only applied in the near out of field regions and should be values between 0.0 and 1.0.

    cell_measured_psf_arrays: List[psf arrays]
         A Sequence containing the M measured PSF arrays that will be interpolated
         to calculate the PSF at any other location in the detector useable area. Each element of the cell array
         is a measured PSF array expressed as a 2D array of size (Dy,Dx)

    array_psf_center_locn: Array[M,3].
        A 2-D array containing two numbers for each of the M measured PSF arrays.
        The three numbers store the center location of the corresponding PSF array in detector coordinates as (Y,X, windowtype).
        The left/lower edge of the first pixel is 0.0, the centre is 0.5 and the right/upper edge is 1.0.

    array_psf_interp_coeffs: Array(NCoeff, NOOFy, NOOFx).
        A 3-D array containing Ncoeffs coefficients used to interpolate the array of measured PSF arrays
        to any other location on the detector surface. Typically Ncoeffs will be ~3 to allow triangular interpolation
        (e.g. Delaunay Triangulation).

    array_psf_interp_indices: Array(NCoeff, NOOFy, NOOFx).
        A 3-D array containing NCoeff integers that index the M measured psf arrays
        in parameter cell_measured_psf_arrays. The indexed arrays are used for the linear interpolation.

    Returns
    -------
     np.ndarray
        The 2d convolution matrix of size Array( Ky*Kx, Ky*Kx). This array forward converts "true" data into the measured
        l1c blurred data. Deconvolution requires the inverse of this matrix.
    """

    print("generate_horizontal_stray_light_matrix: THERE IS A COARSE SAMPLING ISSUE IN THE C++ code THAT NEEDS TO BE FIXED")
    if (vignetting is None) or (vignetting.size == 0):
        vignetting = np.zeros([0, 0], dtype='float64', order='C')

    window_specs = np.array(array_window_specs, dtype='int32', order='C')
    center_locn = np.array(array_psf_center_locn, dtype='float64', order='C')     # Make an array(M, 3)
    assert center_locn.shape[1] == 3
    assert window_specs.size == 22
    s = array_psf_interp_coeffs.shape
    NCoeffs = s[0]
    My = s[1]
    Mx = s[2]
    interp_indices = np.zeros([Mx, My, NCoeffs], dtype='int32', order='C')
    interp_coeffs = np.zeros([Mx, My, NCoeffs], dtype='float64', order='C')

    for i in range(NCoeffs):
        indices = array_psf_interp_indices[i, :, :].reshape([My, Mx])
        coeffs = array_psf_interp_coeffs[i, :, :].reshape([My, Mx])
        interp_indices[:, :, i] = indices.transpose()                                   # M arrays of column major arrays (Nx, Ny)
        interp_coeffs[:, :, i] = coeffs.transpose()                                     # M arrays of column major arrays (Nx, Ny)

    M = len(cell_measured_psf_arrays)                                                   # Get the number of PSF arrays in this list
    s = cell_measured_psf_arrays[0].shape                                               # Get the shape of each PSF array. They must be all the same
    Ny = s[0]                                                                           # Get the Number of rows (Y)
    Nx = s[1]                                                                           # Get the number of columns (X)
    psf_arrays = np.zeros([M, Nx, Ny], dtype='float64', order='C')                      # Create the array for C++ which uses the Column Major transpose
    for i in range(M):                                                                  # and copy
        psf_arrays[i, :, :] = cell_measured_psf_arrays[i].transpose()                   # the transpose for each array over to the system in column major format (Dx, Dy)

    Kx = window_specs[20]
    Ky = window_specs[21]
    A = np.zeros([Kx * Ky, Kx * Ky], dtype='float64', order='C')                        # Create the output array in the column major format
    delta_nx = delta_steps['delta_nx']
    delta_ny = delta_steps['delta_ny']
    delta_sx = delta_steps['delta_sx']
    delta_sy = delta_steps['delta_sy']
    cimpl_set_versionstr(__version__)
    cimpl_create_horizontal_stray_light_matrix(window_specs, vignetting, psf_arrays, center_locn, interp_coeffs, interp_indices, A, delta_nx, delta_ny, delta_sx, delta_sy)
    A = np.array(A.transpose())                                                         # Make a copy back into the row major format (Ny,Nx)
    return A


# ------------------------------------------------------------------------------
#           set_numthreads
# ------------------------------------------------------------------------------
def set_numthreads(numthreads: int) -> bool:
    """
    Configures the altius_l1pp_mex C++ code to use the specified number of threads.

    Parameters
    ----------
    numthreads : int
        Configures altius_l1pp_mex to use this number of threads. A value of 0 or less has no effect.
    """

    ok = cimpl_set_numthreads(int(numthreads))
    return bool(ok)


# ------------------------------------------------------------------------------
#           set_diagnostics
# ------------------------------------------------------------------------------
def set_diagnostics(dodiagnostics: int) -> bool:
    """
    Configures the altius_l1pp_mex C++ code to enable or disable diagnostic modes.

    Parameters
    ----------
    dodiagnostics: bool or int

    Returns
    -------
    bool
        True if success else fail.
    """

    ok = cimpl_set_diagnostic(int(dodiagnostics))
    return bool(ok)
