from typing import Sequence, Tuple, Dict
import xarray as xr
import numpy as np
from datetime import datetime
import os.path


# ------------------------------------------------------------------------------
#           write_to_netcdf
# ------------------------------------------------------------------------------
def write_to_netcdf(filename: str,
                    channel: str,
                    wavelength: float,
                    matrix: np.ndarray,
                    array_window_specs: Sequence[int],
                    array_psf_center_locn: np.ndarray,
                    version: str = "----",
                    description: str = "----"):
    """
    Writes the deconvolution matrix to a netcdf file. This should be compatible with legacy version created by earlier systems.

    Parameters
    ----------
    filename: str
        The full pathname of the netcdf file to be created. It will overwrite the file if it already exists

    channel: str
        A string identifying the channel. It will typically be one of 'uv', 'vis', 'nir'

    wavelength: float
        The wavelength in nanometers of the deconvolution matrix.

    matrix: np.ndarray( Kx*Ky, Kx*Ky)
        The deconvolution matrix. This is the row major 2D square array generated by the deconvolution software. This array
        is written to the netcdf variable ``deconvolution_matrix``.

    array_window_specs : Array[18 or 20].
        Specify the placement and size of the five windows: NOOF, Detector, Science, L1B and L1C

        * [0] = **FULLCONVx**, X size of the full convolution window e.g. 6940 or 1504
        * [1] = **FULLCONVy**, Y size of the full convolution window, eg  6940 or 2000
        * [2] = **NOFSx**, X offset of detector window in the NOOF window, e.g. 2718
        * [3] = **NOFSy**, Y offset of detector window in the NOOF window, e.g. 2718
        * [0] = **NOOFx**, X size of Near-Out-Of-Field window in detector pixels, e.g. 6940
        * [1] = **NOOFy**, Y size of Near-Out-Of-Field window in detector pixels, e.g. 6940
        * [2] = **DOFSx**, X offset of detector window in the NOOF window, e.g. 2718
        * [3] = **DOFSy**, Y offset of detector window in the NOOF window, e.g. 2718
        * [4] = **Dx**, X size of the detector window in detector pixels, e.g. 1504, Dx must equal psf_Nx
        * [5] = **Dy**, Y size of the detector window in detector pixels, e.g. 2000, Dy must equal psf_Ny
        * [6] = **SOFSx**, X offset of science window in the detector window in detector pixels, e.g. 0
        * [7] = **SOFSy**, Y offset of science window in the detector window in detector pixels, e.g. 248
        * [8] = **Sx**, X dimension of science window in detector pixels,  e.g. 1500, Sx must equal Lx*LBINx
        * [9] = **Sy**, Y dimension of science window in detector pixels,  e.g. 1503, sy must equal Ly*LBINy
        * [10] = **LBINx**, X L1B window pixel binning in detector pixels, e.g. 3
        * [11] = **LBINy**, Y L1B window pixel binning in detector pixels, e.g. 3
        * [12] = **Lx**, X number of L1B window pixels e.g. 500. Lx must equal Kx*KBINx
        * [13] = **Ly**, Y number of L1B window pixels e.g. 501. Ly must equal Ky*KBINy
        * [14] = **KBINx**, X L1C window pixel binning in L1B window pixels, e.g. 50
        * [15] = **KBINy**, Y L1C window pixel binning in L1B window pixels, e.g. 1
        * [16] = **Kx**, X number of L1C window pixels, e.g. 10
        * [17] = **Ky**, Y number of L1C window profile pixels, e.g. 501

        This array is written to the netcdf variable ``window_specifications``

    array_psf_center_locn  : Array[M,3].
        A 2-D array containing two numbers for each of the M measured PSF arrays.
        The three numbers store the center location of the corresponding PSF array and the type of coordinates, (Y,X, windowtype).
        ``windowtype`` values are, 2 corresponds to the NOOF window, 3 is the detcetor window and 3 is the science window. Note that
        0.0 corresponds to the left/bottom edge of the first pixel, 0.5 corresponds to the center of the first pixel and 1.0 is the
        right/upper edge of the first pixel. This array is written to the netcdf variable ``psf_location``.


    version: str
        An optional string providing version information. This value is written as as attribute of netcdf variable
        ``deconvolution_matrix``. The default value is "----"

    description: str
        An optional string providing description information. It is recommended that this field be used to describe
        whether the deconvolution matrix is infield, near-out-field etc. This value is written as an attribute of netcdf variable
        ``deconvolution_matrix`` The default value is "----"
    """
    specsname = ['FULLCONVx', 'FULLCONVy',
                 'NOFSx', 'NOFSy', 'NOOFx', 'NOOFy',                                                                                      # Get the standard names of the windwoing parameters
                 'DOFSx', 'DOFSy', 'Dx', 'Dy',
                 'SOFSx', 'SOFSy', 'Sx', 'Sy',
                 'LBINx', 'LBINy', 'Lx', 'Ly',
                 'KBINx', 'KBINy', 'Kx', 'Ky']

    utc = datetime.utcnow().isoformat()                                                                                 # Get the time of file creation
    matrix_array = xr.DataArray(data=matrix,
                                name='deconvolution_matrix',
                                dims=["rows", "columns"],
                                attrs=dict(channel=channel,
                                           description=description,
                                           creation_time='{}'.format(utc),
                                           version=version,
                                           wavelength_nm=wavelength))

    # approximationvar = xr.DataArray(data=int(approximation), name='approximation')                                      # Write the deconvolution approximation value, see SPS usask.sps.L1C.psf.DeconvolutionApproximation

    window_specs = xr.DataArray(data=array_window_specs,                                                                # Write the window specs, it will either be 18 or 20 integers
                                name='window_specifications',
                                dims=["Numspecs"],
                                attrs=dict(description='Window Parameters'))

    assert(array_psf_center_locn.shape[1] == 3)
    psf_location = xr.DataArray(data=array_psf_center_locn,
                                name='psf_central_location',
                                dims=["NumPSF", "Coord"],
                                attrs=dict(description='Central Location of PSF'))
    coordnames = ['y', 'x', 'window_type']

    ds = xr.Dataset(data_vars=dict(deconvolution_matrix=matrix_array,
                                   window_specifications=window_specs,
                                   psf_location=psf_location),
                    coords=dict(specsnames=(["Numspecs"], specsname),
                                coordnames=(["Coord"], coordnames)))

    ds.to_netcdf(filename, encoding=dict(deconvolution_matrix=dict(dtype='float64'),
                                         window_specifications=dict(dtype='int32'),
                                         psf_location=dict(dtype='float64')))


# ------------------------------------------------------------------------------
#         load_from_netcdf
# ------------------------------------------------------------------------------
def load_from_netcdf(filename: str) -> Tuple[np.ndarray, Dict[str, int], np.ndarray]:
    """
    Loads the convolution matrix from a netcdf file. It is assumed this convolution matrix was written using
    method ``write_to_netcdf``

    Parameters
    ----------
    filename: str
        The name of the netcdf file. It must exist, be accessible and be a netcdf file created by method ``write_to_netcdf``.

    Returns
    -------
    Tuple[ convolution_matrix, window_specs, psf_location_array, approximation ]
        Returns a three element tuple. The first element is the convolution matrix stored in the netcdf file. It is returned
        as a numpy array( Ny,Nx). The 2nd element is a dictionary listing the window specifications used to create the convolution
        matrix. The third element is a  numpy array of dimensions (3, Numpsf) that specify the (Y,X,) location of the central lobe of
        each PSF and the window type of the coordinate (2 = NOOF window, 3 = detector window, 4 = Science window).
    """
    if (not os.path.exists(filename)):
        raise ValueError('File <{}> does not exist. It cannot be loaded'.format(filename))
    ds = xr.open_dataset(filename)
    matrix = np.array(ds['deconvolution_matrix'])
    locn = np.array(ds['psf_location'])
    specs = np.array(ds['window_specifications'])
    names = np.array(ds.coords['specsnames'])
    windowspecs = {}
    for i in range(names.size):
        windowspecs[names[i]] = int(specs[i])
    return matrix, windowspecs, locn
